/*
This MemBrain script has been specifically designed to work with time series in combination with both time invariant and 
time variant nets. 
It uses an approach of preparing the time series input data in a way so that the inputs of the net already contain 
a number of past values in a kind of 'ring buffer' way. Through this approach the net itself can be kept time invariant 
since the data patterns themselves already contain the past values of the time series which shall be taken into account. 

Nevertheless, the script can also be used with time variant nets so that a combination of input values from the past as 
well as time variant net architecture can be used.

What the script does:

- Read in time series data from a CSV file generated by the script user.
- Automatically generate a MemBrain lesson from the time series data which has the data points of time t+1 as expected output to patterns from time point t, 
  t-1, t-2, ... and so on.
  I.e. a lesson which is suitable to train nets which shall try to learn the transfer functions between X(t), X(t-1), Y(t-2),... Y(t), Y(t-1), Y(t-2),... --> X(t+1), Y(t+1)...
- Automatically split off a smaller training lesson based on the number of points to predict and the size of the overall data set.
- Train an adjustable number of different net candidates with these data for a specified training time and determine the best candidate with respect to the 
  net error based on the training lesson
- Re-train the best net candidate with the training lesson and predict the next point in time. Record the result of the prediction
- Shift the training data set one point in time towards the future and repeat the training and prediction.
- Repeat the shift and prediction procedure until the specified number of points to predict.
- Merge known correct values and predicted values together in one file and export it to CSV for analysis

The script in principal supports nets of the following architecture with respect their inputs and outputs. You have to design your nets in this way, too, in 
order to be compatible with the script:

INPUTS
I_TS1(t-n),I_TS1(t-n+1),...,I_TS1(t),I_TS2(t-m),I_TS2(t-m+1),I_TS2(t),I_ADD1,I_ADD2,...,I_ADDN

OUTPUTS
O_TS1,O_TS2,...,O_TSN

Nomenclature:
I_TSx(T) = Input neuron for Time Series 'x' at time point 'T'
I_ADDx = Optional additional input neuron 'x' to support the prognosis, not representing a time series
O_TSx = Output neuron for Time Series 'x' at time point 't+1'
n and m define the number of past values which are taken into account and can be defined as script constants separately for each time series within an array in the script. 

In each case you should first examine the files which come packed with the example to understand what the input files and the nets have to look like in principle.
Then you might want to go for your own data, create input files and net candidates for it and use the script to perform data extrapolation into the future with MemBrain.

The input CSV file (TimeSeriesInput.csv) in the example represents a mixture of several sine and cosine waves with different frequencies. One of these waves depends on an 
additional input variable 'Month'. You may want to take a look at the original data and its creation in the spread-sheet that comes with the example package. Its file name is 
'TimeSeriesInput.xlsx'. The spread sheet also contains a diagram which depicts the whole time series including the last 10 data points which MemBrain is finally trying
to predict. In the diagram these data points are coloured differently than the others.
Take a look at the tab 'TimeSeriesInput' in the spread sheet. This tab has been exported to the CSV file 'TimeSeriesInput.csv' for the example. 
 
The script is adjusted to extrapolate 10 points into the future, one at a time, always with previously training on the corresponding preceeding data points.
After the script has executed, the result file 'Extrapolated.csv' is automatically opened in your preferred spread sheet program or text editor - depending on your 
corresponding operating systems settings for opening CSV files. 

To perform the same with your data you need to perform actions as described below. Please note that most of the steps below are one-time steps. 
The only thing you need to do for every data set is the creation of corresponding time series input CSV files.

- Prepare at least one net with inputs x(t-n), x(t-n+1)..., x(t), y(t-m), y(t-m+1)..., y(t), ... and outputs x(t+1), y(t+1)... You can select different names of the 
  neurons if you want to.
  The input neurons from left to right must reflect the same meaning as the output neurons from left to right as depicted above!
  If you use additional inputs (as the 'COS(Month)' in the example) then you have to place them TO THE RIGHT of the other input neurons.
  
Also, don't forget to adjust appropriate normalization ranges for all input and output neurons of your net candidates!

- Copy the files you have created into the directory where the script resides
- In the constant header section of this script adjust the constant values according to your needs.

Note that all your net candidates must have a name that comprises of the value assigned to NET_FILE_NAME_BASE plus a number that starts at '1' for the first net candidate.

For the example, adjust MemBrain as following:
- Edit the teacher 'RPROP' via the Teacher Manager: 
  Disable the option 'Use On-The-Fly Net Error Calculation if possible'. Enable the option 'Reset Net before every Lesson'.
- Select <Teach><Configure Net Error Function>
  On the left hand side select 'Binary by Threshold' and enter the value 0.03
  On the right hand side select 'Mean of Sum'
  
All these are one-time actions only. You can now start the script and it should run seamlessly.

If you want to learn what happens in the script in details try to read through the code. Even if you are not a programmer you should be able to get an 
idea of what is going on according to the comments in the script and the names of the functions that are called. Watch out for the function 'void main()'. 
This is the main program and entry point of the script.
When reading MemBrain script files then best use the Script Debugger that installs together with MemBrain. This brings special syntax highlighting and certainly 
also the possibility to step through the script, inspect variables etc.
*/


// ---------- constants which can be adjusted to tweak the script according to the user's needs -------------------

// The time series input file from which training and validation lesson are generated
const string TIME_SERIES_INPUT_FILE = "TimeSeriesInput.csv";

// Number of neural nets to operate on
const uint8 NUMBER_OF_NETS = 5;

// Number of data points to be extrapolated into the future
const uint EXTRAPOLATE_COUNT = 10;

// Common part of the name of all used neural net candidates. To be extended by number of net 1 .. NUMBER_OF_NETS
const string NET_FILE_NAME_BASE = "Net";

// Default training time per net [s]
const uint32 DEFAULT_TRAIN_TIME = 20;

// Name of combined lesson used for extrapolation
const string EXTRAPOLATION_LESSON_NAME = "ExtrapolationBase.mbl";

// The name with which the final net is stored to disk
const string FINAL_NET_FILE_NAME = "FinalNet.mbn";

// The teacher to be used
const string TEACHER_NAME = "RPROP";

// The extension appended to a net file name when being saved after training
const string NET_FILE_SAVE_AS_EXTENSION = "_Trained";

// File name for the summary
const string RESULT_SUMMARY_FILE_NAME = "TrainingResultSummary.txt";

// Auto-open the file RESULT_SUMMARY_FILE_NAME in the end of the script?
const bool OPEN_RESULT_SUMMARY = false;

// Name of resulting CSV file that contains all result data
const string RESULT_FILE_NAME = "Extrapolated.csv";

// Auto-open the file RESULT_FILE_NAME in the end of the script?
const bool OPEN_RESULT_FILE = true;

// Field separator
const string LIST_SEPARATOR = ";";

// Decimal separator
const string DEC_SEPARATOR = ",";

// Heading for the additionally generated time column in the file <RESULT_FILE_NAME>
const string TIME_COLUMN_HEADING = "#";

// Start value for the additionally generated time column in the file <RESULT_FILE_NAME>
const double TIME_START_VALUE = 0.0;

// Time step size value for the additionally generated time column in the file <RESULT_FILE_NAME>
const double TIME_STEP_SIZE = 1.0;

// Use MemBrain's feature to capture and reload the best net training state from stock
// after the final training cycle? 
const bool USE_BEST_NET_STOCK_CAPTURE = true;

// Shall MemBrain randomize the net again before each training (i.e. for each point extrapolated into the future)?
const bool RANDOMIZE_BEFORE_EACH_EXTRAPOLATION_POINT = false;

// Naming extension added to the output neurons: Output neurons will be named after the corresponding 
// time series name from the imported CSV file plus this extension.
const string OUTPUT_NAMING_EXTENSION = "(t+1)";

// Defines the number of inputs used for each time series
const uint[] INPUTS_PER_TIME_SERIES =
{
	1
	// For more than one time series add a line for each time series followed by a ','. 
	// Don't place a ',' after the last line!	
};

const string TARGET_COLUMN_EXTENSION = "(Target)";		 // Column name extension string for target output values



// ---------- variables with static storage duration below. DON'T CHANGE!!!!!!!!!!!!!! -------------------

// The last known net error for every net
double[] sCurNetErr(NUMBER_OF_NETS);

// The last known minimum net error for every net
double[] sMinNetErr(NUMBER_OF_NETS);

// Index of current net
uint8	sCurNetIdx = 0;

// Training time in seconds
uint32	sTrainTime = 0;

// Has the summary file been opened successfully?
bool sSummaryFileOpen = false;

// The summary file
file sSummaryFile;

// Name of the best net
string sBestNetName;

// To remember the user's settings and restore them after script execution
bool sOldBestNetStockCapture;

// Number of input neurons which are dedicated to time series values 
uint sTimeSeriesInputNeuronCount = 0;

// Pattern number where to currently start training from 
// (will be incremented during operation in order to shift training data set towards the future)
uint sTrainStartPattern = 1;

// Length of the training lesson
uint sTrainLessonSize = 0;


// This is the main script execution function
void main()
{
	// Perform some initialization
	Init();
	
	// Open the first net and open all used windows
	OpenCurrentNet();
	OpenAllWindows();
	
	// Create a training and a validation lesson from the time series input CSV file.
	// The lesson if finally created in lesson #1
	if (!CreateExtrapolationLessonFromTimeSeriesInput())
	{
		AbortScript();
	}
	// Create new exmpy lesson #2
	SetLessonCount(2);
	uint size = GetLessonSize();
	if (size > EXTRAPOLATE_COUNT)
	{
		sTrainLessonSize = size - EXTRAPOLATE_COUNT;
	}
	else
	{
		Trace("Error: Not enough input lines in file " + TIME_SERIES_INPUT_FILE + "\r\n" +
    		  "to extrapolate " + EXTRAPOLATE_COUNT + " into the future!\r\n");
    	AbortScript();	
	}
	
	// Create the first training lesson from lesson #1 to lesson #2
	SplitOffLesson(1, 2, sTrainStartPattern, sTrainLessonSize, false);
		
	// Request some runtime parameters from the user
	RequestTrainingTime();

	// Train and validate all the nets
	for (uint8 i = 0; i < NUMBER_OF_NETS; i++)
	{
		TrainAndValidateNet(i);
	}

	SummarizeResults();
	
	AnnounceAction("Training with rolling lesson and extrapolating " + EXTRAPOLATE_COUNT + " data points into the future (one data point is extrapolated after each training)");
	
	// Now train the best net with a rolling lesson and extrapolate one time point into the future after each training
	// Captured results are located in lesson #3
	TrainAndExtrapolate(sBestNetName, EXTRAPOLATION_LESSON_NAME, sTrainTime, 
                        EXTRAPOLATE_COUNT);
                        
    // Now merge recorded result file with known validation data for easier analysis
    AppendOutputTargetColumns(1, 3, sTrainLessonSize + 1, EXTRAPOLATE_COUNT);
                        		  
    // Add a time column to the left of the inputs to simplify post-processing 
    // in external programs
    AddTimeColumn(3);
    // Export into result CSV file
    ExportLessonRaw(RESULT_FILE_NAME);
                        		  
    Trace("Result data is contained in the file " + RESULT_FILE_NAME + "\r\n");
    Trace("The file contains two data columns for each time series: The first column contains the input data of the training and the validation lesson"
          " (combined into one time series). The second column contains the output data of the net for each input data point.\r\n");
    Trace("Note that the last " + (EXTRAPOLATE_COUNT - 1) + " data points in the input column are the output of the net from the last"
          " point in time!\r\n");
          	
    AutoOpenResultFiles();
   
    // Cleanup operations after script execution
    CleanUp();
}


// Open all windows used during this script.
// Give the user the possibility to properly arrange all the open windows.
void OpenAllWindows()
{
	ShowLessonEditor(false);
	ShowPatternErrorViewer(true);
	ShowErrorViewer(true);
}


// Output a message in the trace window to be the next action of the script
void AnnounceAction(const string &in message)
{
	// Open trace window for sure and leave one empty line after last text entry.
	ShowTraceWin(true);
	Trace(message + "...\n");
}

// Get the file name for the given net index
string GetNetFileFileName(uint8 idx)
{
	return NET_FILE_NAME_BASE + (idx + 1) + ".mbn";
}

// Get the file name for the current net index
string GetNetFileNameCurrentIdx()
{
	return GetNetFileFileName(sCurNetIdx);
}

// Get the file name for the given net ndex to save the trained net as
string GetNetFileNameSaveAs(uint8 idx)
{
	return NET_FILE_NAME_BASE + (idx + 1) + NET_FILE_SAVE_AS_EXTENSION + ".mbn";
}

// Get the file name for the current net ndex to save the trained net as
string GetNetFileNameCurrentIdxSaveAs()
{
	return GetNetFileNameSaveAs(sCurNetIdx);
}


// Open the net corresponding to the current net array index 'sCurNetIdx'
void OpenCurrentNet()
{
	OpenNet(GetNetFileNameCurrentIdx());
}



// Initialize net errors.
void Init()
{
	StopThink();
	StopTeaching();
	
	AnnounceAction("Initializing variables");
	// negative values for net errors indicate invalid values
	for (uint8 i = 0; i < NUMBER_OF_NETS; i++)
	{
		sCurNetErr[i] = -1.0;
		sMinNetErr[i] = -1.0;
	}

	sCurNetIdx = 0;

	AnnounceAction("Selecting teacher");
	SelectTeacher(TEACHER_NAME);
	TeacherSetting(TARGET_ERR, 0.0);
	TeacherSetting(PATTERN_SELECT, ORDERED);
	
	// Remember the user's setting to restore it after script execution
	sOldBestNetStockCapture = GetCaptureBestNetOnStock();
}

// Cleanup operations after script execution
void CleanUp()
{
	// Restore user setting in case we might have modified it
	if (USE_BEST_NET_STOCK_CAPTURE != sOldBestNetStockCapture)
	{
 		SetCaptureBestNetOnStock(sOldBestNetStockCapture);
	}
}


// Request the training time from the user
void RequestTrainingTime()
{
	EDlgRet ret = UserInput("Please enter training time per net in seconds.", DEFAULT_TRAIN_TIME, sTrainTime);

	if (ret != IDOK)
	{
		AbortScript();
	}
}

// Train and validate the net with the given 0-based index
void TrainAndValidateNet(uint8 netIdx)
{
	sCurNetIdx = netIdx;

	string action = "Training and validating net: ";

	action += GetNetFileNameCurrentIdx();
	AnnounceAction(action);

	OpenCurrentNet();
	// Ensure that all nets get the I/O names of the corresponding lesson
	NamesToNet();
	
	RandomizeNet();
	StartTimer(sTrainTime * 1000);

	ETeachResult res = TR_OK;

	StartTeaching();

	// Teach as long as teacher signals OK and as long as timer is not expired.
	// Update net error values during teaching at regular intervals
	while ((res == TR_OK) && !IsTimerExpired() && IsTeaching())
	{
		Sleep(100);
		UpdateNetErrors();

		res = GetLastTeachResult();
	}

	if (IsTeaching())
	{
		// Stop teaching for sure
		StopTeaching();
	}
	UpdateNetErrors();

	SaveNet(GetNetFileNameCurrentIdxSaveAs());
}


// Update the net error variables for later analysis
void UpdateNetErrors()
{
	double err = GetNetError();

	// Store current net error for the given net
	sCurNetErr[sCurNetIdx] = err;

	// Store minimum net error for the given net
	if ((sMinNetErr[sCurNetIdx] < 0) || (err < sMinNetErr[sCurNetIdx]))
	{
		sMinNetErr[sCurNetIdx] = err;
	}
}


// Provide a result summary of the training
void SummarizeResults()
{
	Trace("\n------------------ Training result summary start ----------------------\n\n");

	string result;

	// Try to open result summary file
	sSummaryFileOpen = sSummaryFile.Open(RESULT_SUMMARY_FILE_NAME, FILE_MODE_WRITE | FILE_MODE_CREATE | FILE_SHARE_DENY_NONE);

	if (!sSummaryFileOpen)
	{
		MessageBox("Unable to open summary file!");
	}

	// Create headline in summary file
	Append2SummaryFile("Net" + LIST_SEPARATOR + "Last Net Error" + LIST_SEPARATOR + "Min Net Error\n");

	double minNetErr = -1;
	uint8 bestNetIdx = 0;

	for (uint8 i = 0; i < NUMBER_OF_NETS; i++)
	{
		result = GetNetFileNameSaveAs(i);
		Trace(result + ":\n");
		Append2SummaryFile(result + LIST_SEPARATOR);

		result = sCurNetErr[i];
		result = ConvertDoubleStringToLocale(result);
		Trace("Last known net error: " + result + ":\n");
		Append2SummaryFile(result + LIST_SEPARATOR);

		result = sMinNetErr[i];
		result = ConvertDoubleStringToLocale(result);
		Trace("Minimum net error: " + result + ":\n\n");
		Append2SummaryFile(result + "\n");

		if ((minNetErr < 0) || (sCurNetErr[i] < minNetErr))
		{
			minNetErr = sCurNetErr[i];
			bestNetIdx = i;
		}
	}

	sBestNetName = GetNetFileNameSaveAs(bestNetIdx);
	Trace("The best trained net is '" + sBestNetName + "'\n");

	Trace("\n------------------ Training result summary end ----------------------\n\n");

	Append2SummaryFile("\nBest trained net: " + sBestNetName + "\n");

	if (sSummaryFileOpen)
	{
		Trace("All results have been additionally summarized in the file '" + RESULT_SUMMARY_FILE_NAME + "'.\n");
		sSummaryFile.Close();
	}

}


// Append text to the summary file if open
void Append2SummaryFile(string text)
{
	if (sSummaryFileOpen)
	{
		if (!sSummaryFile.Write(text))
		{
			MessageBox("Error in writing to summary file!");
			sSummaryFile.Close();
			sSummaryFileOpen = false;
		}
	}
}


// Convert the given number string to incorporate local decimal separator
const string ConvertDoubleStringToLocale(const string &in numStr)
{
	if (DEC_SEPARATOR == ".")
	{ // Nothing to do if anglo-american notation is adjusted
		return numStr;
	}

	int idx = numStr.findFirst(".", 0);

	if (idx < 0)
	{ // Nothing to do if string does not contain separator
		return numStr;
	}

	// The string contains a separator. Replace by user defined separator now...

	// copy part left from separator
	string ret = numStr.substr(0, idx);

	// Add user defined separator
	ret += DEC_SEPARATOR;

	// copy part right from separator
	ret += numStr.substr(idx + 1, numStr.length() - idx - 1);

	return ret;
}


// Train the net.
void Train(uint trainTime)
{
    if (RANDOMIZE_BEFORE_EACH_EXTRAPOLATION_POINT)
    {
    	RandomizeNet();
    }
    
    // Adjust the option to capture best net during training on stock
    SetCaptureBestNetOnStock(USE_BEST_NET_STOCK_CAPTURE);
    
    StartTimer(trainTime * 1000);

	ETeachResult res = TR_OK;

	StartTeaching();

	// Teach as long as teacher signals OK and as long as timer is not expired.
	// Update net error values during teaching at regular intervals
	while ((res == TR_OK) && !IsTimerExpired() && IsTeaching())
	{
		Sleep(100);
		
        res = GetLastTeachResult();
	}

	if (IsTeaching())
	{
		// Stop teaching for sure
		StopTeaching();
	}
	
	// Reload the best training version of this net from stock if desired
	if (USE_BEST_NET_STOCK_CAPTURE)
	{
		string fileName;
		GetCurrentNetFileName(fileName);
		uint slot = GetLastTeacherAutoCaptureStockSlot();
		string message = "Loading best training version of net ";
		message += fileName + "\r\n";
		message += "from stock slot ";
		message += slot + "...\r\n";
		AnnounceAction(message);
		LoadNetFromStock(slot);
	}
}



// Final training and extrapolation extrapolate into the future
// Resulting (extrapolated) data is located in lesson #3 after the function returns 
void TrainAndExtrapolate(string &in netFile, string &in lessonFile, uint trainTime, 
                        uint extrapolateCount)
{
    bool ok = true;
    
    // Open net, delete all lessons except for #1, create an additional, empty lesson #2
    // import training data into lesson #1
    OpenNet(netFile);
    SaveNet(FINAL_NET_FILE_NAME);
    RandomizeNet();
    SetLessonCount(1);
    // Load training lesson
    SelectLesson(1);
    LoadLesson(lessonFile);
    // Create new empty lesson for training and for result file
    SetLessonCount(3);
    SetRecordingType(RT_ACT);
    
    // disable view update during think
    ViewSetting(UPDATE_THINK, false);
        
    for (uint i = 0; i < EXTRAPOLATE_COUNT; i++)
    {
    	Trace("Processing data point #" + (i + 1) + "...\r\n");
    	// Create the next training lesson from lesson #1 to lesson #2
    	SplitOffLesson(1, 2, sTrainStartPattern, sTrainLessonSize, true);
		SelectLesson(2);
		// The net error shall be calculated based on the training lesson
    	SelectNetErrLesson(2);
    	Train(trainTime);
    	// Now perform one think step on the pattern following the last training pattern in the original
    	// lesson and record the result to lesson #3
    	EnableLessonEditorUpdate(false);
    	SelectLesson(1);
    	SelectPattern(sTrainStartPattern + sTrainLessonSize);
    	ApplyPattern();
    	StartRecording(3);
    	ThinkSteps(1);
    	SleepExec();
    	StopRecording();
    	sTrainStartPattern++;
    	SelectLesson(2);
    	EnableLessonEditorUpdate(true);
    }
    
    // Re-enable view update during think
    ViewSetting(UPDATE_THINK, true);
}




// Add a time column to the left in the specified lesson number
void AddTimeColumn(uint lessonNum)
{
	// Select the recorded lesson
	SelectLesson(lessonNum);
	
	// Disable Lesson Editor screen update for higher performance
	EnableLessonEditorUpdate(false);
	
	// We need one more input column. This will by default be generated to the right of the other columns
    uint oldInCount = GetLessonInputCount();
    SetLessonInputCount(oldInCount + 1);
    // Since we want our time column on the left hand side we need to move all columns in the input section 
    // one position to the right, including the names of the columns
    
    // First move all names
    string name;
    for (uint col = oldInCount; col >= 1; col--)
    {
    	GetLessonInputName(col, name);
    	SetLessonInputName(col + 1, name);
    }  
    
    // Now move all data rows
    uint rows = GetLessonSize();
    double value;
    for (uint row = 1; row <= rows; row++)
    {
    	SelectPattern(row);
    	for (uint col = oldInCount; col >= 1; col--)
    	{
    		GetPatternInput(col, value);
    		SetPatternInput(col + 1, value);
    	}  
    }   
    
    
    // Set the heading of the new column on the left side
    SetLessonInputName(1, TIME_COLUMN_HEADING);
    // Generate time axis data
    double time = TIME_START_VALUE;
    
    for (uint row = 1; row <= rows; row++)
    {
    	SelectPattern(row);
    	SetPatternInput(1, time);
    	time += TIME_STEP_SIZE;	
    }
	
	EnableLessonEditorUpdate(true);
}


// Automatically open the result files as configured in the script constants section
void AutoOpenResultFiles()
{
	if (OPEN_RESULT_SUMMARY)
	{
		ShellExecute("open", RESULT_SUMMARY_FILE_NAME, "");
	}
	
	if (OPEN_RESULT_FILE)
	{
		ShellExecute("open", RESULT_FILE_NAME, "");
	}
}

// Create a combined extrapolation lesson from the time series input CSV file.
// Data from time points t-n, t-n+1, ... will be combined with data from time points t+1 
// to build input data / expected output data.
bool CreateExtrapolationLessonFromTimeSeriesInput()
{
	Trace("Create extrapolation lesson...\r\n");
	// Import the time series input file into lesson #1 according to the format of
 	// the loaded net
 	SetLessonCount(1);
 	ClearLesson();
	NamesFromNet();
	// The number of different time series to be trained is reflected in the loaded net
	// by the number of outputs 
    uint timeSeriesCount = GetOutputCount();
    uint inCount = GetInputCount();
    uint timeSeriesLength = 0;
    double value;
    uint maxInputsPerTimeSeries = 1;
    
    // Disable Lesson Editor screen update for higher performance
	EnableLessonEditorUpdate(false);
	ShowPatternErrorViewer(false);
    
    // The net must have the correct number of inputs according to the number of used data points
    // for each time series and optional additional input data columns.
    // Note that all additional inputs must be placed to the right of the time series inputs!
        
    // First, we need a definition of the number of used inputs for each time series 
    uint temp = INPUTS_PER_TIME_SERIES.length();
    bool ok = (timeSeriesCount == INPUTS_PER_TIME_SERIES.length());
    
    if (ok)
    {
    	for (uint i = 0; i < timeSeriesCount; i++)
    	{
    		uint inputs = INPUTS_PER_TIME_SERIES[i];
    		sTimeSeriesInputNeuronCount += inputs;
    		if (inputs > maxInputsPerTimeSeries)
    		{
    			maxInputsPerTimeSeries = inputs;
    		}
    	}
    	
    	ok = (inCount >= sTimeSeriesInputNeuronCount);
    	if (ok)
    	{
    		uint addInputCount = inCount - sTimeSeriesInputNeuronCount;
    		// In order to correctly import the time series input file we need as many columns as there
    		// are time series plus the number of additional input columns which are not time series.
    		SetLessonInputCount(timeSeriesCount + addInputCount);
    		// The time series input file does not have any output columns 
    		ImportLessonInputsRaw(TIME_SERIES_INPUT_FILE);
			// Determine length of time series (number of data points per time serie)
			timeSeriesLength = GetLessonSize();
			// A time series must consist of at least two data points and of at least enough data points to
			// for one valid data pattern
			ok = (timeSeriesLength > maxInputsPerTimeSeries);
			if (ok)
			{
				SetLessonCount(2);
				CreateExtrapolationLesson(1, 2, maxInputsPerTimeSeries, sTimeSeriesInputNeuronCount);
			}
			else
    		{
    			Trace("Input time series must contain at least " + maxInputsPerTimeSeries + " data rows!\r\n");
    		}
		}
		else
		{
			Trace("At least " + sTimeSeriesInputNeuronCount + " input neurons expected according to the number of outputs and the \r\n"
			      "definition of the script constant 'INPUTS_PER_TIME_SERIES'\r\n"); 
		}
    }
    else
    {
    	Trace("The script array 'INPUTS_PER_TIME_SERIES' must provide exactly one value for each time series\r\n"
    		 "(i.e. one for each output neuron)!\r\n");
    }
    
    if (ok)
    {
    	// Delete lesson #2. Lesson #1 will become selected automatically
    	SetLessonCount(1);
    	// Load the extrapolation lesson to lesson #1. This is our overall working basis from which we
    	// split off the single training lessons and from which we take the known target data.
    	LoadLesson(EXTRAPOLATION_LESSON_NAME);
    }
    
    EnableLessonEditorUpdate(true);
    ShowPatternErrorViewer(true);
    PatternErrorViewerFitYScale();
	
	return ok;
}

// Create a combined extrapolation lesson from the time series input CSV file.
// Time series data is already imported into lesson <sourceLessonNum>.
// The target lesson <targetLessonNum> has already been allocated.
// Data from time points t-n, t-n+1, ... will be combined with data from time points t+1 
// to build input data / expected output data.
void CreateExtrapolationLesson(uint sourceLessonNum, uint targetLessonNum, uint maxInputsPerTimeSeries, uint timeSeriesInputCount)
{
	SelectLesson(sourceLessonNum);
	
	uint timeSeriesLength = GetLessonSize();
	uint sourceLessonInCount = GetLessonInputCount();
	uint timeSeriesCount = GetOutputCount();
	
	
	SelectLesson(targetLessonNum);
	NamesFromNet();
	// Create meaningful names for the outputs according to the inputs
    string name;
    uint inputPos = 1;
    
    EnableLessonEditorUpdate(false);
    
    for (uint col = 1; col <= timeSeriesCount; col++)
    {
    	SelectLesson(sourceLessonNum);
    	GetLessonInputName(col, name);
    	SelectLesson(targetLessonNum);
    	SetLessonOutputName(col, name + OUTPUT_NAMING_EXTENSION);
    	
    	uint timePointCount = INPUTS_PER_TIME_SERIES[col - 1];
    	
    	for (uint i = timePointCount; i > 0; i--)
    	{
    		if (i == 1)
    		{
    			SetLessonInputName(inputPos, name + "(t)");
    		}
    		else
    		{
    			SetLessonInputName(inputPos, name + "(t-" + (i - 1) + ")");
    		}
    		inputPos++;
    	}
    }
    
    EnableLessonEditorUpdate(true);
    // Assign the lesson names to the net
    NamesToNet();
    EnableLessonEditorUpdate(false);
    
    // Now perform the time shift: For each pattern (i.e. input data row) copy 
	// the corresponding data fields from the next row as expected output data 
	// (data point t+1 will become the expected output of data points t).
	// For the last input row there is no expected output. We have to delete this input
	// row.
	double value;
	uint sourceRow = 1;
	double[] sourceInputVector;
			
	sourceInputVector.resize(sourceLessonInCount);
		
	// Pre-fill all inputs with past values and generate first training pattern once all 
	// input values are filled.
	// Create and first pattern in the target lesson
    SelectLesson(targetLessonNum);
    AddPattern();
    for (sourceRow = 1; sourceRow <= timeSeriesLength; sourceRow++)
    {
    	SelectLesson(sourceLessonNum);
    	SelectPattern(sourceRow);
    	
    	// Read in whole input row to simplify processing later on
    	for (uint inputNum = 1; inputNum <= sourceLessonInCount; inputNum++)
    	{
    		GetPatternInput(inputNum, value);
    		sourceInputVector[inputNum - 1] = value;
    	}
    	// Select target pattern
    	SelectLesson(targetLessonNum);
        SelectPattern(GetLessonSize());
    	
    	if (sourceRow > maxInputsPerTimeSeries)
    	{ // All inputs pre-filled. Add output values from next input lesson pattern now and exit this preparation 
    	  // loop
    		for (uint inputNum = 1; inputNum <= timeSeriesCount; inputNum++)
    		{
    			SetPatternOutput(inputNum, sourceInputVector[inputNum - 1]);
    		}
    		break;
    	}
    	
    	for (uint inputNum = 1; inputNum <= sourceLessonInCount; inputNum++)
    	{
    		if (inputNum > timeSeriesCount)
    		{ // Auxiliary input: Just copy to target lesson as inputs to the right of the time series inputs
    			uint targetCol = timeSeriesInputCount + (inputNum - timeSeriesCount);
    			SetPatternInput(targetCol, sourceInputVector[inputNum - 1]);	
    		} 
    		else
    		{ // Time series input
    			uint pastPoints = INPUTS_PER_TIME_SERIES[inputNum - 1];
    			if (sourceRow > maxInputsPerTimeSeries - pastPoints)
    			{
    				uint targetCol = 1;
    				for (uint i = 0; i < inputNum - 1; i++)
    				{
    					targetCol += INPUTS_PER_TIME_SERIES[i];
    				}
    				targetCol += (sourceRow - (maxInputsPerTimeSeries - pastPoints) - 1); 
    				SetPatternInput(targetCol, sourceInputVector[inputNum - 1]);
    			} 
    		}
    	}	
    }
    // First pattern created now.
    sourceRow++;
    // Create all following patterns now using time shift within each time series input field:
    // 1.) Shift inputs to the left by one
    // 2.) Move pattern outputs to the inputs (time point t, right-most of each input field)
    // 3.) Take input values from time series input and set them as pattern outputs
    double[] targetTimeSeriesInputVector;
    double[] targetTimeSeriesOutputVector;
    			
	targetTimeSeriesInputVector.resize(timeSeriesInputCount);
	targetTimeSeriesOutputVector.resize(timeSeriesCount);
    
    for (; sourceRow <= timeSeriesLength; sourceRow++)
    {
    	SelectLesson(sourceLessonNum);
    	SelectPattern(sourceRow);
    	// Read in whole input row to simplify processing later on
    	for (uint inputNum = 1; inputNum <= sourceLessonInCount; inputNum++)
    	{
    		GetPatternInput(inputNum, value);
    		sourceInputVector[inputNum - 1] = value;
    	}
    	SelectLesson(targetLessonNum);
    	// Read in all target lesson time series inputs to simplify processing later on
    	for (uint inputNum = 1; inputNum <= timeSeriesInputCount; inputNum++)
    	{
    		GetPatternInput(inputNum, value);
    		targetTimeSeriesInputVector[inputNum - 1] = value;
    	}
    	// Read in all target lesson time series outputs to simplify processing later on
    	for (uint outputNum = 1; outputNum <= timeSeriesCount; outputNum++)
    	{
    		GetPatternOutput(outputNum, value);
    		targetTimeSeriesOutputVector[outputNum - 1] = value;
    	}
    	AddPattern();
    	SelectPattern(GetLessonSize());
    	
    	for (uint inputNum = 1; inputNum <= sourceLessonInCount; inputNum++)
    	{
    		if (inputNum > timeSeriesCount)
    		{ // Auxiliary input: Just copy to target lesson as inputs to the right of the time series inputs
    			uint targetCol = timeSeriesInputCount + (inputNum - timeSeriesCount);
    			SetPatternInput(targetCol, sourceInputVector[inputNum - 1]);	
    		} 
    		else
    		{ // Time series input
    			uint offset = 1;
    			for (uint i = 0; i < inputNum - 1; i++)
    			{
    				offset += INPUTS_PER_TIME_SERIES[i];
    			}
    			uint pastPoints = INPUTS_PER_TIME_SERIES[inputNum - 1];
    			for (uint i = 0; i < pastPoints - 1; i++)
    			{ // Time shift to the left
    				value = targetTimeSeriesInputVector[offset + i];
    				SetPatternInput(offset + i, value);
    			} 
    			// Output copied to the last input time slot
    			value = targetTimeSeriesOutputVector[inputNum - 1];
    			SetPatternInput(offset + pastPoints - 1, value);
    		}
    	}
    	
    	// Add output values from input lesson pattern now
    	for (uint inputNum = 1; inputNum <= timeSeriesCount; inputNum++)
    	{
    		SetPatternOutput(inputNum, sourceInputVector[inputNum - 1]);
    	}	
    }
    
    EnableLessonEditorUpdate(true);
    
    // The just created lesson with all data points combined of t and t+1 is our
    // overall time series lesson. Save it to file now.
    SaveLesson(EXTRAPOLATION_LESSON_NAME);
}


// Split off a lesson from the source lesson, starting at the given pattern number
// and with the given length
void SplitOffLesson(uint sourceLesson, uint targetLesson, uint startPattern, uint length, bool useShiftOperation)
{
	array<double> dummy;
	SelectLesson(targetLesson);
	
	if (useShiftOperation && (GetLessonSize() > 0))
	{ // ONly perform a shift operation by one, no need to copy everything again
		// Delete the firsat pattern
		SelectPattern(1);
		DeletePattern();
		// Append the last pattern from the target lesson
		CopyPattern(sourceLesson, startPattern + length - 1, targetLesson, dummy);
	}
	else
	{ // Fully copy operation
		EnableLessonEditorUpdate(false); // To improve performance
		ShowPatternErrorViewer(false); // Improves speed
		ClearLesson();
		
		for (uint i = 0; i < length; i++)
		{
			CopyPattern(sourceLesson, startPattern + i, targetLesson, dummy);
		} 
		EnableLessonEditorUpdate(true);
		ShowPatternErrorViewer(true); // Improves speed
	}	
	PatternErrorViewerFitYScale();
}


// Read an input vector from the Lesson Editor to a RAM array 
// (as per currently selected lesson and pattern)
void GetCurrentPatternInputs(array<double> &out inputs)
{
	uint inCount = GetLessonInputCount();
	double value;
		
	inputs.resize(inCount);
	for (uint i = 0; i < inCount; i++)
	{
		GetPatternInput(i + 1, value);
		inputs[i] = value;
	}
}

// Set an input vector in the Lesson Editor from a RAM array 
// (as per currently selected lesson and pattern)
void SetCurrentPatternInputs(array<double> inputs)
{
	uint inCount = GetLessonInputCount();
	
	for (uint i = 0; i < inCount; i++)
	{
		SetPatternInput(i + 1, inputs[i]);
	}
}

// Read an output vector from the Lesson Editor to a RAM array 
// (as per currently selected lesson and pattern)
void GetCurrentPatternOutputs(array<double> &out outputs)
{
	uint outCount = GetLessonOutputCount();
	double value;
		
	outputs.resize(outCount);
	for (uint i = 0; i < outCount; i++)
	{
		GetPatternOutput(i + 1, value);
		outputs[i] = value;
	}
}

// Set an output vector in the Lesson Editor from a RAM array 
// (as per currently selected lesson and pattern)
void SetCurrentPatternOutputs(array<double> outputs)
{
	uint outCount = GetLessonOutputCount();
	
	for (uint i = 0; i < outCount; i++)
	{
		SetPatternOutput(i + 1, outputs[i]);
	}
}



// Copy a given pattern from a given source lesson to a given target lesson
// Also fill a certain number of input column values to the left including from an array
void CopyPattern(uint sourceLesson, uint patNo, uint targetLesson, array<double> &in leadingInputColValues)
{
	SelectLesson(targetLesson);
	AddPattern();
	
	uint targetCount = GetLessonSize();
	
	SelectPattern(targetCount);
	SelectLesson(sourceLesson);
	SelectPattern(patNo);
	
	array<double> inputs;
	array<double> outputs;
	
	GetCurrentPatternInputs(inputs);
	
	uint leadingInputColCount = leadingInputColValues.length();
	
	if (leadingInputColCount > 0)
	{
		for (uint i = 0; i < leadingInputColCount; i++)
		{
			inputs.insertAt(0, leadingInputColValues[leadingInputColCount - 1 - i]);
		}
	} 
	
	GetCurrentPatternOutputs(outputs);
	SelectLesson(targetLesson);
	SetCurrentPatternInputs(inputs);
	SetCurrentPatternOutputs(outputs);
}

// Append the original (target value) output columns to the recorded validation lesson
void AppendOutputTargetColumns(uint valiLesson, uint recordedLesson, uint startPattern, uint patternCount)
{
	EnableLessonEditorUpdate(false);
	
	SelectLesson(valiLesson);
	SelectPattern(startPattern);
	uint addedCount = GetLessonOutputCount();
	SelectLesson(recordedLesson);
	uint oldOutCount = GetLessonOutputCount();
	SetLessonOutputCount(oldOutCount + addedCount);
	double act;
	string name;
	
	// Copy column names + name extention
	for (uint col = 1; col <= addedCount; col++)
	{  
		SelectLesson(valiLesson);
		GetLessonOutputName(col, name);
		SelectLesson(recordedLesson);
		SetLessonOutputName(oldOutCount + col, name + " " + TARGET_COLUMN_EXTENSION);
	}
	
	// Copy data
	for (uint pat = 1; pat <= patternCount; pat++)
	{
		for (uint col = 1; col <= addedCount; col++)
		{  
			SelectLesson(valiLesson);
			SelectPattern(startPattern + pat - 1);
			GetPatternOutput(col, act);
			SelectLesson(recordedLesson);
			SelectPattern(pat);
			SetPatternOutput(oldOutCount + col, act);
		}
	}
	
	EnableLessonEditorUpdate(true);
}

// Automatically called when script aborts (e.g. error)
void OnAbortScript()
{
	StopTeaching();
	//MessageBox("The script was aborted before having finished!");
}
